; ultra high-speed memory allocation system
;
; suitable for lots (2000+) of small allocations and releases of the same size
; minimum size: 4 bytes
;	size must be a multiple of four
; zero memory overhead, zero fragmentation
;
; if you free some space, it is reserved for later use
;
; use fsAlloc to allocate space
;	use fsFree to free space (don't forget to supply the size too!)
; use fsReleaseAll to release all memory used by this allocator for the given size (other sizes are untouched)

ifndef FSALLOC_H
FSALLOC_H equ 1

; uses ecx, edx
; return value in eax
fsAlloc macro _size:req	; always use the same notation for your size
	LOCAL _new, _newspace, _done
	
	ifndef fsFree_&_size
		.data
		align 4
		fsFree_&_size		dd	0
		fsAddr_&_size		dd	0
		fsSpace_&_size	dd	0
		fsMem_&_size		dd	0
		.code
	endif
	
	cmp fsFree_&_size, 0
	jz _new
	
	mov eax, fsFree_&_size
	mov ecx, [eax]
	mov fsFree_&_size, ecx
	jmp _done
	
_new:
	cmp fsSpace_&_size, _size
	jl _newspace
	mov eax, fsAddr_&_size
	add fsAddr_&_size, _size
	sub fsSpace_&_size, _size
	jmp _done
	
_newspace:
;  if (_size <= 16)
;		push PAGE_EXECUTE_READWRITE
;		push MEM_COMMIT
;		push _size*16384
;		push NULL
;		call VirtualAlloc
;		mov ecx, fsMem_&_size
;		mov [eax], ecx
;		mov fsMem_&_size, eax
;		add eax, 4
;		lea edx, [eax + _size]
;		mov fsAddr_&_size, edx
;		mov fsSpace_&_size, _size*(16384-1) - 4
;	elseif (_size <= 64)
;		push PAGE_EXECUTE_READWRITE
;		push MEM_COMMIT
;		push _size*4096
;		push NULL
;		call VirtualAlloc
;		mov ecx, fsMem_&_size
;		mov [eax], ecx
;		mov fsMem_&_size, eax
;		add eax, 4
;		lea edx, [eax + _size]
;		mov fsAddr_&_size, edx
;		mov fsSpace_&_size, _size*(4096-1) - 4
;	elseif ((%_size) <= 1024)
		push PAGE_EXECUTE_READWRITE
		push MEM_COMMIT
		push _size*1024
		push NULL
		call VirtualAlloc
		mov ecx, fsMem_&_size
		mov [eax], ecx
		mov fsMem_&_size, eax
		add eax, 4
		lea edx, [eax + _size]
		mov fsAddr_&_size, edx
		mov fsSpace_&_size, _size*(1024-1) - 4
;	else
;		push PAGE_EXECUTE_READWRITE
;		push MEM_COMMIT       
;		push _size*64
;		push NULL
;		call VirtualAlloc
;		mov ecx, fsMem_&_size
;		mov [eax], ecx
;		mov fsMem_&_size, eax
;		add eax, 4
;		lea edx, [eax + _size]
;		mov fsAddr_&_size, edx
;		mov fsSpace_&_size, _size*(1024-1) - 4
;	endif
	
_done:
endm

; uses eax, edx
fsFree macro mem:req, _size:req
	mov eax, mem
	mov edx, fsFree_&_size
	mov [eax], edx
	mov fsFree_&_size, eax
endm

; uses eax, ecx, edx
fsReleaseAll macro _size:req
	LOCAL _loop
	
	push esi
	mov esi, fsMem_&_size
	
_loop:
	cmp esi, 0
	jz _done
	
	push MEM_RELEASE
	push 0
	push esi
	mov esi, [esi]
	call VirtualFree
	
	jmp _loop
_done:
	mov fsFree_&_size, 0
	mov fsAddr_&_size, 0
	mov fsSpace_&_size, 0
	mov fsMem_&_size, 0
	pop esi
endm

endif